"""
Predefined kernels used for convolution by arcgis.raster.functions.convolution()
in addition to user defined kernels
"""

LINE_DETECTION_HORIZONTAL = 0
LINE_DETECTION_VERTICAL = 1
LINE_DETECTION_LEFT_DIAGONAL = 2
LINE_DETECTION_RIGHT_DIAGONAL = 3
GRADIENT_NORTH = 4
GRADIENT_WEST = 5
GRADIENT_EAST = 6
GRADIENT_SOUTH = 7
GRADIENT_NORTH_EAST = 8
GRADIENT_NORTH_WEST = 9
SMOOTH_ARITHMETIC_MEAN = 10
SMOOTHING_3X3 = 11
SMOOTHING_5X5 = 12
SHARPENING_3X3 = 13
SHARPENING_5X5 = 14
LAPLACIAN_3X3 = 15
LAPLACIAN_5X5 = 16
SOBEL_HORIZONTAL = 17
SOBEL_VERTICAL = 18
SHARPEN = 19
SHARPEN2 = 20
POINT_SPREAD = 21



"""
The arcgis.widgets module provides components for visualizing GIS data and analysis.
This module includes the MapView Jupyter notebook widget for visualizing maps and layers
"""
import json
import random
import string
from uuid import uuid4

from arcgis.features import FeatureSet, Feature, FeatureCollection
from arcgis.raster import ImageryLayer
from arcgis.gis import Layer
from arcgis.gis import Item
from arcgis.mapping import WebMap
from arcgis._impl.common._utils import _date_handler
from arcgis.geometry import Point, Polygon, Polyline, MultiPoint, Geometry
try:
    from ipywidgets import widgets
except:
    from IPython.html import widgets

try:
    from traitlets import Unicode, Int, List, Bool, Dict
except:
    from IPython.utils.traitlets import Unicode, Int, List, Bool, Dict

import logging

_LOGGER = logging.getLogger(__name__)


class MapView(widgets.DOMWidget):
    """Mapping widget for Jupyter Notebook"""

    #region Class, instance and interop variables
    _view_name = Unicode('MapView').tag(sync=True)
    _view_module = Unicode('mapview').tag(sync=True)

    # value = Unicode('Hello World!').tag(sync=True)

    basemap = Unicode('topo').tag(sync=True)
    _basemap = Unicode('topo').tag(sync=True)
    basemaps = List([]).tag(sync=True)
    _gallerybasemaps = List([]).tag(sync=True)
    _gbasemaps_def = List([]).tag(sync=True)
    _js_basemap = Unicode('').tag(sync=True)
    _js_renderer = Unicode('').tag(sync=True)
    _js_layer_list = Unicode('').tag(sync=True)  # sync layer - to keep track of layers represented in JS code
    _js_interactive_drawn_graphic = Unicode('').tag(sync=True)  # get geometries of interactively drawn graphics
    _layerId_to_remove = Unicode('').tag(sync=True)
    _widget_layer_list = [] #internal to keep track of layers added, removed from widget
    _widget_graphics_list = [] #internal to keep track of layers sent to draw() of the widget
    _webmap_added_layer_list = []
    _webmap_drawn_layer_list = []
    width = Unicode('100%').tag(sync=True)
    zoom = Int(2).tag(sync=True)
    id = Unicode('').tag(sync=True)
    center = List([0, 0]).tag(sync=True)
    mode = Unicode('navigate').tag(sync=True)
    _addlayer = Unicode('').tag(sync=True)
    start_time = Unicode('').tag(sync=True)
    end_time = Unicode('').tag(sync=True)
    _extent = Unicode('').tag(sync=True)
    _jsextent = Unicode('').tag(sync=True)
    _token_info = Unicode('').tag(sync=True)
    _arcgis_url = Unicode('').tag(sync=True)
    _swipe_div = Unicode('').tag(sync=True)
    _gallery_initialized = False # Used to keep track if GalleryBasemaps has been called
    #endregion

    def __init__(self, **kwargs):
        """Constructor of Map widget.
        Accepts the following keyword arguments:
        gis     The gis instance with which the map widget works, used for authentication, and adding secure layers and
                private items from that GIS
        item    web map item from portal with which to initialize the map widget
        """
        super(MapView, self).__init__(**kwargs)
        self._click_handlers = widgets.CallbackDispatcher()
        self._draw_end_handlers = widgets.CallbackDispatcher()

        self.on_msg(self._handle_map_msg)

        self.basemaps = ['dark-gray',
                        'dark-gray-vector',
                        'gray',
                        'gray-vector',
                        'hybrid',
                        'national-geographic',
                        'oceans',
                        'osm',
                        'satellite',
                        'streets',
                        'streets-navigation-vector',
                        'streets-night-vector',
                        'streets-relief-vector',
                        'streets-vector',
                        'terrain',
                        'topo',
                        'topo-vector']
        self._swipe_div = 'swipeDiv' + ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(6))

        self._gis = kwargs.pop('gis', None)
        if self._gis is not None and self._gis._con._username is not None:  # not anonymous
            token_info = {
                "server": self._gis._con.baseurl.replace('http://', 'https://'),
                "tokenurl": (self._gis._con.baseurl +
                             'generateToken').replace('http://', 'https://'),
                "username": self._gis._con._username,
                "password": self._gis._con._password
            }
            self._token_info = json.dumps(token_info)
            # if self._gis.properties.portalName != 'ArcGIS Online':
            self._arcgis_url = self._gis._con.baseurl + 'content/items'

        self.item = kwargs.pop('item', None)

        #construct a webmap in memory
        self._webmap = WebMap()
        self._widget_layer_list = []  # initialize to empty list.
        self._widget_graphics_list = []  # initialize to empty list for each widget.
        self._webmap_added_layer_list = []  # keep track of layers in internal webmap from add_layer()
        self._webmap_drawn_layer_list = []  # keep track of layers in internal webmap from draw()
        self._smartmapping_layer_list = []  # keep track of layers that are rendered using smart mapping.

        if self.item is not None:
            if isinstance(self.item, WebMap):
                self.item = self.item.item

                # constuct the internal webmap using the passed web map item
                self._webmap = WebMap(self.item)

                # use internal webmap layers as widget layers
                self._widget_layer_list = self._webmap.layers
                for l in self._webmap.layers:
                    self._webmap_added_layer_list.append(l.id)  # append layer ids to added layer list.

            if 'type' in self.item and self.item.type.lower() != 'web map':
                raise TypeError("item type must be web map")
            self.id = self.item.id

    def draw(self, shape, popup=None, symbol=None, attributes=None):
        """
        Draws a shape on the map widget. You can draw anything from known geometries, coordinate pairs, FeatureSet
        objects.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        shape                  Required object.

                               Known geometries:
                               shape is one of ["circle", "downarrow", "ellipse", "extent", "freehandpolygon",
                                                "freehandpolyline", "leftarrow", "line", "multipoint", "point",
                                                "polygon", "polyline", "rectangle", "rightarrow", "triangle",
                                                "uparrow", or geometry dict object]

                               Coordinate pair: specify shape as a list of [lat, long]. Eg: [34, -81]

                               FeatureSet: shape can be a FeatureSet object.
        ------------------     --------------------------------------------------------------------
        popup                  Optional dict. Dict containing "title" and "content" as keys that will be displayed
                               when the shape is clicked. In case of a FeatureSet, "title" and "content" are names of
                               attributes of the features in the FeatureSet instead of actual string values for title
                               and content.
        ------------------     --------------------------------------------------------------------
        symbol                 Optional dict. symbol is specified in json format as described at
                               http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000000n5000000. A
                               default symbol is used if one is not specified.

                               Tip: a helper utility to get the symbol format for several predefined symbols is
                               available at http://esri.github.io/arcgis-python-api/tools/symbol.html
        ------------------     --------------------------------------------------------------------
        attributes             Optional dict. Specify a dict containing name value pairs of fields and field values
                               associated with the graphic.
        ==================     ====================================================================
        """

        title = attributes['title'] if attributes and 'title' in attributes else "Notebook sketch layer"

        if isinstance(shape, list) and len(shape) == 2:  # [lat, long] pair
            shape = {'x': shape[1], 'y': shape[0], "spatialReference": {"wkid": 4326}, 'type': 'point'}

        elif isinstance(shape, tuple):  # (lat, long) pair
            shape = {'x': shape[1], 'y': shape[0], "spatialReference": {"wkid": 4326}, 'type': 'point'}

        elif isinstance(shape, dict) and 'location' in shape: # geocoded location
            shape = {'x': shape['location']['x'], 'y': shape['location']['y'],
                     "spatialReference": {"wkid": 4326}, 'type': 'point'}

        if isinstance(shape, FeatureSet):
            fset = shape
            for feature in fset.features:
                if popup:
                    popup_dict = {'title':feature.attributes[popup['title']],
                                  'content':feature.attributes[popup['content']]}
                else:
                    popup_dict = None
                graphic = {
                    "geometry": feature.geometry,
                    "infoTemplate": popup_dict,
                    "symbol": symbol,
                    "attributes": feature.attributes
                }
                self.mode = json.dumps(graphic)

            #add as layer to internal webmap
            if popup:
                webmap_popup = {'title':"{" +popup['title']+"}",
                                'description':"{" + popup['content'] +"}"}
            else:
                webmap_popup = None

            # add at save time for consistency.
            self._webmap_drawn_layer_list.append({'data':shape,
                                                  'options':{'popup':webmap_popup, 'symbol':symbol, 'attributes':attributes,
                                                            'title':title, 'extent':self.extent}})

        elif isinstance(shape, dict):
            graphic = {
                "geometry": shape,
                "infoTemplate": popup,
                "symbol": symbol,
                "attributes": attributes
            }
            self.mode = json.dumps(graphic)
            # print(json.dumps(graphic))
            # build feature, then featureSet
            try:
                f = Feature(shape)
                fset = FeatureSet([f], geometry_type='esriGeometryPoint', spatial_reference={'wkid':4326})

                if popup:
                    webmap_popup = {'title':popup['title'], 'description':popup['content']}
                else:
                    webmap_popup = None

                # add at save time for consistency.
                self._webmap_drawn_layer_list.append({'data':fset,
                                                      'options':{'popup':webmap_popup, 'symbol':symbol, 'attributes':attributes,
                                              'title':title, 'extent':self.extent}})
            except:
                pass

        else:  # interactive sketches on map widget.
            self.mode = shape

    def add_layer(self, item, options=None):
        """
        Adds the specified layer or item to the map widget.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        item                   Required object. You can specify Item objects, Layer objects such as
                               FeatureLayer, ImageryLayer, MapImageLayer etc., FeatureSet and
                               FeatureCollection objects.
        ------------------     --------------------------------------------------------------------
        options                Optional dict. Specify visualization options such as renderer info,
                               opacity, definition expressions. See example below
        ==================     ====================================================================

        .. code-block:: python  (optional)

           USAGE EXAMPLE: Add a feature layer with smart mapping renderer and a definition expression to limit the
           features drawn.

           map1 = gis.map("Seattle, WA")
           map1.add_layer(wa_streets_feature_layer, {'renderer':'ClassedSizeRenderer',
                                                     'filed_name':'DistMiles',
                                                     'opacity':0.75})
        """

        self._js_renderer = ''
        if isinstance(item, list):
            for lyr in item:
                self.add_layer(lyr, options)

        elif isinstance(item, dict) and 'function_chain' in item:
            js_layer = item['layer']._lyr_json
            options_dict = {
                    "imageServiceParameters": {
                        "renderingRule": item['function_chain']
                    }
                }

            if options is not None:
                options_dict.update(options)

            js_layer.update({
                "options": json.dumps(options_dict)
            })
            self._addlayer = json.dumps(js_layer)
            #add to widget's layer list
            self._widget_layer_list.append(item)

        elif isinstance(item, FeatureSet):
            fc = FeatureCollection.from_featureset(item)
            self.add_layer(fc, options)

        elif isinstance(item, Layer):
            if isinstance(item, FeatureCollection):  # apply symbol, title for FeatureCollections
                if not options:
                    options = {'title':'Notebook sketch layer'}
                if options and 'title' not in options:
                    options['title'] = 'Notebook sketch layer'
                if options and 'symbol' in options:
                    if hasattr(item.layer, 'layers'):
                        item.layer.layers[0].layerDefinition.drawingInfo['renderer']['symbol'] = options['symbol']
                    else:
                        item.layer.layerDefinition.drawingInfo['renderer']['symbol'] = options['symbol']

            js_layer = item._lyr_json

            if 'type' in js_layer:
                if js_layer['type'] == 'MapImageLayer':
                    js_layer['type'] = 'ArcGISTiledMapServiceLayer' if 'TilesOnly' in item.properties.capabilities else 'ArcGISDynamicMapServiceLayer'

            if options is not None:
                if 'options' in js_layer:  # ImageryLayers may have rendering rules in options
                    lyr_options = json.loads(js_layer['options'])
                    lyr_options.update(options)
                    js_layer.update({'options': json.dumps(lyr_options)})
                else:
                    js_layer.update({"options": json.dumps(options)})
            else:
                options = {} #to store extent and other properties
            if 'uses_gbl' in js_layer:
                if js_layer["uses_gbl"] is True:
                    _LOGGER.warning("""Imagery layer object containing global functions in the function chain cannot be used for dynamic visualization.
                                   \nThe layer output must be saved as a new image service before it can be visualized. Use save() method of the layer object to create the processed output.""")
                    return None
            self._addlayer = json.dumps(js_layer,
                                        default=_date_handler)
            options['extent'] = self.extent

            self._webmap.add_layer(item, options)
            # add to widget's layer list
            self._widget_layer_list.append(item)

            # if smart mapping, then add it to the internal list
            if options and 'renderer' in options:
                self._smartmapping_layer_list.append({item: options})

        elif 'layers' in item:  # items as well as services
            if item.layers is None:
                raise RuntimeError('No layers accessible/available in this item or service')
            for lyr in item.layers:
                self.add_layer(lyr, options)  # recurse.
                # js_layer = lyr._lyr_json
                # if options is not None:
                #     js_layer.update({"options": json.dumps(options)})
                # self._addlayer = json.dumps(js_layer)
                #
                # self._webmap.add_layer(lyr, options)
                # # add to widget's layer list
                # self._widget_layer_list.append(item)

        else:  # dict {'url':'xxx', 'type':'yyy', 'opacity':'zzz' ...}
            if options is not None:
                item.update({"options": json.dumps(options)})

            self._addlayer = json.dumps(item)

    def clear_graphics(self):
        """
        Clear the graphics drawn on the map widget. Graphics are shapes drawn using the 'draw()' method.
        :return:
        """
        self.mode = "###clear_graphics"
        self._js_interactive_drawn_graphic = ""
        self._webmap_drawn_layer_list = []

    def set_time_extent(self, start_time, end_time):
        self.start_time = start_time
        self.end_time = end_time

    def remove_layers(self, layers=None):
        """
        Removes the layers added to the map widget. You can get the list of layers added to the widget by querying the
        'layers' property.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        layers                 Optional list. Specify the list of layers to be removed from the map widget. You can get
                               the list of layers on the map by querying the 'layers' property.

                               If not specified, it removes all layers added to the widget.
        ==================     ====================================================================
        :return:
            True if layer is successfully removed. Else, False.
        """

        if not layers:
            # empty layer list, then remove all layers
            if self.layers:
                while len(self.layers) > 0:
                    self.remove_layers(layers=self.layers)
                return True
            else:
                return True

        if 'layers' in layers:  # if you passed in an item
            self.remove_layers(layers=layers.layers)

        elif isinstance(layers, list):
            # removal is async. Hence, loop until there is no layer left.
            for layer in layers:
                # find index - common across all internal lists
                index_to_remove = self._widget_layer_list.index(layer)

                # region remove from widget
                try:
                    js_layers_parsed = json.loads(self._js_layer_list)
                    js_layer_to_remove = js_layers_parsed[index_to_remove]

                    self._layerId_to_remove = js_layer_to_remove['id']

                    js_layers_parsed.remove(js_layer_to_remove)

                    # remove the layer from synced trait
                    self._js_layer_list = json.dumps(js_layers_parsed)
                except:
                    # layer to remove does not exist in JS list or, JS list is empty (can happen in case of duplicate
                    # layers, or unreachable services).
                    pass

                # endregion

                # region remove from webmap
                webmap_layer_to_remove = self._webmap.layers[index_to_remove]
                self._webmap.remove_layer(webmap_layer_to_remove)
                # endregion

                # region remove from widget layer list
                self._widget_layer_list.remove(layer)
                # endregion
            return True

        else:
            return False

    @property
    def gallery_basemaps(self):
        """
        Retrieves the portal's custom basemap group and populates properties
        """

        if not self._gallery_initialized:
        #if len(self._gallerybasemaps) == 0:
            bmlyrs = []
            bms = []
            bmquery = self._gis.properties['basemapGalleryGroupQuery']
            basemapsgrp = self._gis.groups.search(bmquery, outside_org=True)
            if len(basemapsgrp) == 1:
                for bm in basemapsgrp[0].content():
                    if bm.type.lower() == 'web map': # Only use WebMaps
                        bms.append(bm.title.lower().replace(" ", "_"))  # item title will be name
                        item_data = bm.get_data()  # we have to get JSON definition to pass through
                        if item_data is not None:
                            bmlyrs.append(item_data['baseMap']['baseMapLayers'])
                #self._gbasemaps_def = bmlyrs
                self._gbasemaps_def = self._gbasemaps_def + bmlyrs
                #nm = self._gis.properties['defaultBasemap']['title']
                #print("Loading Gallery Basemaps....")
                #self._gallerybasemaps = bms
                self._gallerybasemaps = self._gallerybasemaps + bms
                self._gallery_initialized = True
                return self._gallerybasemaps
            else:
                #print("Basemap Group '" + str(bmquery) + "' could not be found...")
                #return [] # If unable to find the group, return empty list
                return self._gallerybasemaps # Return whatever state List is in, even if empty
        else:
            return self._gallerybasemaps

    @property
    def basemap(self):
        """
        Name of the basemap displayed on the map widget. To change the basemap, simply assign it a different value.
        You can get the list of all basemap values by calling the `basemaps` property.
        :return:
        """
        return self._basemap

    @basemap.setter
    def basemap(self, value):
        if isinstance(value, str):
            if (value in self.basemaps):
                self._basemap = value
            elif (value in self.gallery_basemaps): # this should initialize
                self._basemap = value
            else:
                print("Basemap '" + str(value) + "' is not a valid basemap name.")
        self._update_webmap_basemap()

    def _update_webmap_basemap(self):
        """
        Internal method, reads the mapwidget's basemap and applied that for the internal web map
        :return:
        """

        if self._js_basemap:
            basemap_urls = json.loads(self._js_basemap)
            basemap_layers = []

            for layer in basemap_urls:
                #construct basemap layer
                basemap_layers_dict = {'id': uuid4().__str__(),
                                       'visibility': True,
                                       'opacity': 1,
                                       'title': layer['title']}

                if 'vector' in layer['title']:
                    basemap_layers_dict['layerType'] = "VectorTileLayer"
                    basemap_layers_dict['styleUrl'] = layer['url']
                else:
                    basemap_layers_dict['layerType'] = "ArcGISTiledMapServiceLayer"
                    basemap_layers_dict['url'] = layer['url']

                #append to list of basemap layer
                basemap_layers.append(basemap_layers_dict)

            #update basemap in internal web map
            self._webmap._basemap['baseMapLayers'] = basemap_layers
            self._webmap._basemap['title'] = layer['title']
            self._webmap._webmapdict['baseMap'] = self._webmap._basemap


        # Allow for a WebMap item to be passed in and set.  This is useful if someone
        #  finds some other WebMap in AGOL/Portal that they would like to use in this
        #  notebook.  In this case, you need to add it to the current set.
        #  This will utilize the gallery_basemaps objects for adding the new basemap.
        #elif (isinstance(value, Item)):
        #    if value.type.lower() == "web map":
        #        webmapitem = value
        #        bmlyrs = []
        #        bms = []
        #        bms.append(webmapitem.title.lower().replace(" ", "_")) # item title will be name
        #        item_data = webmapitem.get_data()  # we have to get JSON definition to pass through
        #        if item_data is not None:
        #            bmlyrs.append(item_data['baseMap']['baseMapLayers'])
        #            self._gbasemaps_def = self._gbasemaps_def + bmlyrs
        #            self._gallerybasemaps = self._gallerybasemaps + bms
        #            self._basemap = bms[0]
        #            print('Set Map Widget Basemap to WebMap Item basemap layers.')
        #        else:
        #            print("WebMap item appears to have no item_data.")

    @property
    def extent(self):
        if self._jsextent is not None and self._jsextent != '':  # first preference to _jsextent
            return json.loads(self._jsextent)
        elif self._extent:
            ext = json.loads(self._extent)
            if not 'spatialReference' in ext:
                ext['spatialReference'] = {'wkid':4326, 'latestWkid':4326}
            return ext

    @extent.setter
    def extent(self, value):
        if isinstance(value, (tuple, list)):
            if all(isinstance(el, list) for el in value):
                extent = {
                    'xmin': value[0][0],
                    'ymin': value[0][1],
                    'xmax': value[1][0],
                    'ymax': value[1][1]
                }
                value = extent
        self._extent = json.dumps(value)

    @property
    def layers(self):
        """
        The list of layers added to the map widget using the `add_layers()` method.
        :return:
        """
        return self._widget_layer_list

    def on_click(self, callback, remove=False):
        """Register a callback to execute when the map is clicked.

        The callback will be called with one argument,
        the clicked widget instance.

        Parameters
        ----------
        remove : bool (optional)
            Set to true to remove the callback from the list of callbacks."""
        self._click_handlers.register_callback(callback, remove=remove)

    def on_draw_end(self, callback, remove=False):
        """Register a callback to execute when something is drawn

        The callback will be called with two argument,
        the clicked widget instance, and the geometry drawn

        Parameters
        ----------
        remove : bool (optional)
            Set to true to remove the callback from the list of callbacks."""
        self._draw_end_handlers.register_callback(callback, remove=remove)
        #todo - get geometry and call add_layer of webmap and sync the layer lists.

    def _handle_map_msg(self, _, content, buffers):
        """Handle a msg from the front-end.

        Parameters
        ----------
        content: dict
            Content of the msg."""

        if content.get('event', '') == 'mouseclick':
            self._click_handlers(self, content.get('message', None))
        if content.get('event', '') == 'draw-end':
            self._draw_end_handlers(self, content.get('message', None))

    def _process_drawn_graphics(self):
        """internal method to add graphics layers to web map.
        This needs to be done at the end since drawing on widget is async"""
        if self._webmap_drawn_layer_list:
            for element in self._webmap_drawn_layer_list:
                self._webmap.add_layer(element['data'], element['options'])

        if self._js_interactive_drawn_graphic:
            # geometry types could be one of [polyline, polygon, multipoint]
            drawn_graphic = json.loads(self._js_interactive_drawn_graphic)
            if drawn_graphic:
                for element in drawn_graphic:
                    # construct a Geometry object
                    if element['geometry']['type'] == 'polyline':
                        geom = Polyline(element['geometry'])
                    elif element['geometry']['type'] == 'polygon':
                        geom = Polygon(element['geometry'])
                    elif element['geometry']['type'] == 'multipoint':
                        geom = MultiPoint(element['geometry'])
                    elif element['geometry']['type']=='point':
                        geom = Point(element['geometry'])
                    else:
                        geom = Geometry(element['geometry'])

                    # construct a Feature and FeatureSet
                    feat = Feature(geom)
                    fset = FeatureSet([feat])

                    self._webmap.add_layer(fset, {'title':'Map notes'})

    def _process_smart_mapping(self):
        """internal method to insert smart mapping renderer info if used"""
        js_layers_parsed = json.loads(self._js_layer_list)
        for items in self._smartmapping_layer_list:
            lyr = list(items.keys())[0]
            opt = list(items.values())[0]
            # find the index in web map
            try:
                sm_layer_index = self._widget_layer_list.index(lyr)
                js_sm_layer = js_layers_parsed[sm_layer_index]
                wm_sm_layer = self._webmap._webmapdict['operationalLayers'][sm_layer_index]

                # region - heatmap renderer
                if opt['renderer'] in ['HeatmapRenderer', 'ClassedColorRenderer', 'ClassedSizeRenderer']:
                    rend_dict = js_sm_layer['renderer']

                    # update web map renderer
                    if 'featureCollection' not in wm_sm_layer:  # regular web layer
                        wm_sm_layer['layerDefinition']['drawingInfo'] = {'renderer': rend_dict}
                    else:
                        wm_sm_layer['featureCollection']['layers'][0]['layerDefinition']['drawingInfo'] = \
                            {'renderer': rend_dict}
                # endregion

            except ValueError:
                continue  # that layer with smart mapping must have been removed at a later time

    def save(self, item_properties, thumbnail=None, metadata=None, owner=None, folder=None):
        """
        Save the map widget object into a new web map Item in your GIS.

        .. note::
            If you started out with a fresh map widget object, use this method to save it as a the web map item in your
            GIS.

            If you started with a map widget object from an existing web map item or WebMap object, calling this method
            will create a web map new item with your changes. If you want to update the existing web map item with your
            changes, call the `update()` method instead.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_properties     Required dictionary. See table below for the keys and values.
        ---------------     --------------------------------------------------------------------
        thumbnail           Optional string. Either a path or URL to a thumbnail image.
        ---------------     --------------------------------------------------------------------
        metadata            Optional string. Either a path or URL to the metadata.
        ---------------     --------------------------------------------------------------------
        owner               Optional string. User object corresponding to the desired owner of this
                            item. Defaults to the logged in user.
        ---------------     --------------------------------------------------------------------
        folder              Optional string. Name of the folder where placing item.
        ===============     ====================================================================

        *Key:Value Dictionary Options for Argument item_properties*

        =================  =====================================================================
        **Key**            **Value**
        -----------------  ---------------------------------------------------------------------
        typeKeywords       Optional string. Provide a lists all sub-types, see URL 1 below for valid values.
        -----------------  ---------------------------------------------------------------------
        description        Optional string. Description of the item.
        -----------------  ---------------------------------------------------------------------
        title              Optional string. Name label of the item.
        -----------------  ---------------------------------------------------------------------
        tags               Optional string. Tags listed as comma-separated values, or a list of strings.
                           Used for searches on items.
        -----------------  ---------------------------------------------------------------------
        snippet            Optional string. Provide a short summary (limit to max 250 characters) of the what the item is.
        -----------------  ---------------------------------------------------------------------
        accessInformation  Optional string. Information on the source of the content.
        -----------------  ---------------------------------------------------------------------
        licenseInfo        Optional string.  Any license information or restrictions regarding the content.
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Locale, country and language information.
        -----------------  ---------------------------------------------------------------------
        access             Optional string. Valid values are private, shared, org, or public.
        -----------------  ---------------------------------------------------------------------
        commentsEnabled    Optional boolean. Default is true, controls whether comments are allowed (true)
                           or not allowed (false).
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Language and country information.
        =================  =====================================================================

        URL 1: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000000ms000000

        :return:
            Item object corresponding to the new web map Item created.

        .. code-block:: python

           USAGE EXAMPLE: Save map widget as a new web map item in GIS

           map1 = gis.map("Italy")
           map1.add_layer(Italy_streets_item)
           map1.basemap = 'dark-gray'
           italy_streets_map = map1.save({'title':'Italy streets',
                                        'snippet':'Arterial road network of Italy',
                                        'tags':'streets, network, roads'})
        """

        # update the map extent
        self._webmap._extent = self.extent

        # add drawn graphics to web map
        self._process_drawn_graphics()

        # process smart mapping
        if self._smartmapping_layer_list:
            self._process_smart_mapping()

        # update the basemap
        self._update_webmap_basemap()

        return self._webmap.save(item_properties, thumbnail, metadata, owner, folder)

    def update(self, item_properties=None, thumbnail=None, metadata=None):
        """
        Updates the web map item that was used to create the MapWidget object. In addition, you can update
        other item properties, thumbnail and metadata.

        .. note::
            If you started out a MapView object from an existing web map item, use this method to update the web map
            item in your with your changes.

            If you started out with a fresh MapView object (without a web map item), calling this method will raise
            a RuntimeError exception. If you want to save the map widget into a new web map item, call the `save()`
            method instead.

            For item_properties, pass in arguments for only the properties you want to be updated.
            All other properties will be untouched.  For example, if you want to update only the
            item's description, then only provide the description argument in item_properties.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_properties     Optional dictionary. See table below for the keys and values.
        ---------------     --------------------------------------------------------------------
        thumbnail           Optional string. Either a path or URL to a thumbnail image.
        ---------------     --------------------------------------------------------------------
        metadata            Optional string. Either a path or URL to the metadata.
        ===============     ====================================================================

        *Key:Value Dictionary Options for Argument item_properties*

        =================  =====================================================================
        **Key**            **Value**
        -----------------  ---------------------------------------------------------------------
        typeKeywords       Optional string. Provide a lists all sub-types, see URL 1 below for valid values.
        -----------------  ---------------------------------------------------------------------
        description        Optional string. Description of the item.
        -----------------  ---------------------------------------------------------------------
        title              Optional string. Name label of the item.
        -----------------  ---------------------------------------------------------------------
        tags               Optional string. Tags listed as comma-separated values, or a list of strings.
                           Used for searches on items.
        -----------------  ---------------------------------------------------------------------
        snippet            Optional string. Provide a short summary (limit to max 250 characters) of the what the item is.
        -----------------  ---------------------------------------------------------------------
        accessInformation  Optional string. Information on the source of the content.
        -----------------  ---------------------------------------------------------------------
        licenseInfo        Optional string.  Any license information or restrictions regarding the content.
        -----------------  ---------------------------------------------------------------------
        culture            Optional string. Locale, country and language information.
        -----------------  ---------------------------------------------------------------------
        access             Optional string. Valid values are private, shared, org, or public.
        -----------------  ---------------------------------------------------------------------
        commentsEnabled    Optional boolean. Default is true, controls whether comments are allowed (true)
                           or not allowed (false).
        =================  =====================================================================

        URL 1: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#//02r3000000ms000000

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

           USAGE EXAMPLE: Interactively add a new layer and change the basemap of an existing web map.

           italy_streets_item = gis.content.search("Italy streets", "Web Map")[0]
           map1 = MapView(item = italy_streets_item)

           map1.add_layer(Italy_streets2)
           map1.basemap = 'dark-gray-vector'
           map1.update(thumbnail = './new_webmap.png')
        """

        if not self.item:
            raise RuntimeError('Item object missing, you should use `save()` method if you are creating a '
                               'new web map item')

        # update the map extent
        self._webmap._extent = self.extent

        # add drawn graphics to web map
        self._process_drawn_graphics()

        # process smart mapping
        if self._smartmapping_layer_list:
            self._process_smart_mapping()

        # update the basemap
        self._update_webmap_basemap()

        return self._webmap.update(item_properties, thumbnail, metadata)


"""
Allows access to the Portal Logs
"""
from .. import GIS
from ._base import BasePortalAdmin
########################################################################
class Logs(BasePortalAdmin):
    """
    Logs are records written by various components of the portal. You can
    query the logs, clean the logs, and edit log settings.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    gis               required GIS, portal connection object
    ----------------  ---------------------------------------------------------------
    url               required string, web address of the log resource
    ================  ===============================================================

    """
    _gis = None
    _url = None
    _con = None
    _portal = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        if isinstance(gis, GIS):
            self._url = url
            self._gis = gis
            self._portal = gis._portal
            self._con = gis._con
        else:
            raise ValueError("gis object must be of type GIS")
    #----------------------------------------------------------------------
    def clean(self):
        """
        Deletes all the log files on the machine hosting Portal for ArcGIS.
        This operation allows you to free up disk space. The logs cannot be
        recovered after executing this operation.
        """
        url = "%s/clean" % self._url
        params = {"f" : "json"}
        res = self._con.post(path=url, postdata=params)
        if isinstance(res, dict) and \
           'status' in res:
            return res['status'] == "success"
        return False
    #----------------------------------------------------------------------
    @property
    def settings(self):
        """
        Reads/writes the current log settings for the portal.
        """
        url = "%s/settings" % self._url
        params = {'f' : 'json'}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    @settings.setter
    def settings(self, value):
        """
        Reads/writes the current log settings for the portal.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        value             required dictionary, the dictionary of the log settings
        ================  ===============================================================

        :returns:
           None
        """
        url = "%s/settings/edit" % self._url
        params = {'f' : 'json'}
        if isinstance(value, dict):
            for k,v in value.items():
                params[k] = v
        else:
            raise ValueError("Value must be a dictionary")
        return self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def query(self,
              start_time,
              end_time=None,
              level="WARNING",
              query_filter="*",
              page_size=1000):
        """
        The query operation allows you to aggregate, filter, and page
        through logs written by the portal.

        ================  ===============================================================
        **Argument**      **Description**
        ----------------  ---------------------------------------------------------------
        start_time        required datetime/float. The most recent time to query.

                          Local date corresponding to the POSIX timestamp, such as is
                          returned by time.time(). This may raise OverflowError, if the
                          timestamp is out of the range of values supported by the
                          platform. It's common for this to be restricted to years from
                          1970 through 2038.
                          Time can be specified as a portal timestamp (format in
                          "%Y-%m-%dT%H:%M:%S") or in seconds since UNIX epoch. For
                          :Examples:
                          Datetime Object: datetime.datetime.now()
                          Timestamp: "2015-08-01T15:17:20,123"
                          Seconds: 1312237040.123/time.time()
                          Default: datetime.datetime.now()
        ----------------  ---------------------------------------------------------------
        end_time          optional datetime/float, The oldest time to include in the
                          result set. You can use this to limit the query to the last
                          number of minutes, hours, days, months, and years as needed.

                          Local date corresponding to the POSIX timestamp, such as is
                          returned by time.time(). This may raise OverflowError, if the
                          timestamp is out of the range of values supported by the
                          platform. It's common for this to be restricted to years from
                          1970 through 2038.

                          Datetime Object: datetime.datetime.now()
                          Timestamp: "2015-08-01T15:17:20,123"
                          Seconds: 1312237040.123/time.time()
                          Default: datetime.datetime.now()
        ----------------  ---------------------------------------------------------------
        level             optional string, Can be one of [OFF, SEVERE, WARNING, INFO,
                          FINE, VERBOSE, DEBUG]. Returns only records with a log level at
                          or more severe than the level specified.
                          Default: WARNING
        ----------------  ---------------------------------------------------------------
        query_filter      optional string, Filtering is allowed by any combination of
                          codes, users, and source components. The filter accepts a comma
                          delimited list of filter definitions. If any definition is
                          omitted, it defaults to all ("*").
                          :Example:
                          {"codes":[204000-205999,212015,219114], "users":["admin","jcho"],
                           "source": ["PORTAL ADMIN"]}
                          The source of logged events are generated from the sharing,
                          administrative, and portal components of the software.
                          For example:
                           - Events related to publishing and users are categorized under
                             SHARING.
                           - Events related to security and indexing are categorized under
                             PORTAL ADMIN.
                           - Events related to installing the software are categorized
                             under PORTAL.
        ----------------  ---------------------------------------------------------------
        page_size         optional integer, the number of log records to return. The
                          default is 1000
        ================  ===============================================================

        :returns:
           dictionary of messages
        """
        from datetime import datetime
        from six import integer_types, string_types
        url = "%s/query" % self._url
        if isinstance(start_time, datetime):
            start_time = start_time.strftime("%Y-%m-%dT%H:%M:%S")
        elif isinstance(start_time, string_types):
            try:
                datetime.strptime(start_time, "%Y-%m-%dT%H:%M:%S")
            except:
                raise Exception("Invalid start_time string, must be in the format YYYY-MM-DDTHH:MM:SS")
        elif isinstance(start_time, tuple(list(integer_types) + [float])):
            start_time = datetime.utcfromtimestamp(start_time).strftime("%Y-%m-%dT%H:%M:%S")
        if end_time is None:
            end_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        elif isinstance(end_time, datetime):
            end_time = end_time.strftime("%Y-%m-%dT%H:%M:%S")
        elif isinstance(end_time, string_types):
            try:
                datetime.strptime(end_time, "%Y-%m-%dT%H:%M:%S")
            except:
                raise Exception("Invalid end_time string, must be in the format YYYY-MM-DDTHH:MM:SS")
        elif isinstance(end_time, tuple(list(integer_types) + [float])):
            end_time = datetime.utcfromtimestamp(end_time).strftime("%Y-%m-%dT%H:%M:%S")
        if query_filter == "*":
            query_filter = {"codes":[], "users":[], "source": "*"}
        params = {
            "startTime" : start_time,
            "endTime" : end_time,
            "level" : level,
            "f" : "json",
            "filterType" : "json",
            "pageSize" : page_size

        }
        if query_filter:
            params['filter'] = query_filter
        return self._con.get(path=url, params=params)



try:
    import arcpy
    hasArcPy = True
except:
    hasArcPy = False
import tempfile
#----------------------------------------------------------------------
def scratchFolder():
    if hasArcPy:
        return arcpy.env.scratchFolder
    else:
        return tempfile.gettempdir()
#----------------------------------------------------------------------
def scratchGDB():
    if hasArcPy:
        return arcpy.env.scratchGDB
    return tempfile.gettempdir()
#----------------------------------------------------------------------
def json_to_featureclass(json_file,
                         out_fc):
    if hasArcPy:
        return arcpy.JSONToFeatures_conversion(in_json_file=json_file,
                                               out_features=out_fc)[0]
    return None
#----------------------------------------------------------------------



""" Defines util functions used interanally by the workforce-python-api.
"""

import datetime


def from_arcgis_date(timestamp):
    """ Converts an ArcGIS timestamp to a datetime.
        :param timestamp: An int representing an ArcGIS timestamp
        :return: An equivalent datetime.datetime.
    """
    if timestamp is not None:
        return datetime.datetime.fromtimestamp(timestamp / 1000, datetime.timezone.utc)
    return None


def to_arcgis_date(datetime_obj):
    """ Converts a datetime to an ArcGIS timestamp.
        :param datetime_obj: A datetime.datetime
        :return: An int expressing the same date/time as an ArcGIS timestamp.
    """
    if datetime_obj is not None:
        return datetime_obj.timestamp() * 1000
    return None



""" Defines store functions for working with Tracks.
"""
from ... import workforce
from .utils import remove_features, validate, update_features, add_features


def get_track(project, object_id=None, global_id=None):
    """ Gets the identified track.  Exactly one form of identification should be provided.
        :param project:
        :param object_id: The track's OBJECTID.
        :param global_id: The track's GlobalID.
        :returns: Track
    """
    if object_id is not None:
        where = "{}={}".format(project._track_schema.object_id, object_id)
    elif global_id is not None:
        where = "{}='{}'".format(project._track_schema.global_id, global_id)
    else:
        where = "1=0"
    tracks = query_tracks(project, where)
    return tracks[0] if tracks else None


def get_tracks(project):
    """ Gets all tracks in the project.
        :param project:
        :returns: list of Tracks
    """
    return query_tracks(project, '1=1')


def add_tracks(project, tracks):
    """ Adds Tracks to a project.

        Side effect: Upon successful addition on the server, the object_id and global_id fields of
        each Track in tracks will be updated to the values assigned by the server.

        :param project:
        :param tracks: list of tracks
        :raises ValidationError: Indicates that one or more tracks failed validation.
        :raises ServerError: Indicates that the server rejected the tracks.
    """
    if tracks:
        use_global_ids = True
        for track in tracks:
            validate(track._validate_for_add)

            if track.global_id is None:
                use_global_ids = False

        features = [track.feature for track in tracks]
        add_features(project.tracks_layer, features, use_global_ids)
    return tracks


def add_track(project, feature=None, geometry=None, accuracy=None):
    """ Adds a track to the project.

        Side effect: Upon successful addition on the server, the object_id and global_id fields of
        each Track in tracks will be updated to the values assigned by the server.

        :param project:
        :param feature: the feature
        :param geometry: the geometry
        :param accuracy: the accuracy of the track
        :raises ValidationError: Indicates that one or more tracks failed validation.
        :raises ServerError: Indicates that the server rejected the tracks.
    """
    track = workforce.Track(project, feature, geometry, accuracy)
    return add_tracks(project, [track])[0]


def update_tracks(project, tracks):
    """ Updates Tracks.
        :param project:
        :param tracks: list of tracks to update
        :raises ValidationError: Indicates that one or more tracks failed validation.
        :raises ServerError: Indicates that the server rejected the tracks.
    """
    if tracks:
        for track in tracks:
            validate(track._validate_for_update)
        features = [track.feature for track in tracks]
        update_features(project.tracks_layer, features)
    return tracks


def update_track(project, track, geometry=None, accuracy=None):
    if accuracy:
        track.accuracy = accuracy
    if geometry:
        track.geometry = geometry
    return update_tracks(project, [track])[0]


def query_tracks(project, where):
    """ Executes a query against the tracks feature layer.
        :param project: The project in which to query tracks.
        :param where: An ArcGIS where clause.
        :returns: list of Tracks
    """
    track_features = project.tracks_layer.query(where, return_all_freatures=True).features
    return [workforce.Track(project, feature) for feature in track_features]


def delete_tracks(project, tracks):
    """ Removes tracks from the project.
        :param project:
        :param tracks: list of Tracks
        :raises ValidationError: Indicates that one or more tracks failed validation.
        :raises ServerError: Indicates that the server rejected the removal.
    """
    if tracks:
        for track in tracks:
            validate(track._validate_for_remove)
        remove_features(project.tracks_layer, [track.feature for track in tracks])



"""
The Summarize Data module contains functions that calculate total counts, lengths, areas, and basic descriptive statistics of features and their attributes within areas or near other features.

aggregate_points calculates statistics about points that fall within specified areas or bins.
join_features calculates statistics about features that share a spatial, temporal, or attribute relationship with other features.
reconstruct_tracks calculates statistics about points or polygons that belong to the same track and reconstructs inputs into tracks.
summarize_attributes calculates statistics about feature or tabular data that share attributes.
summarize_within calculates statistics for area features and attributes that overlap each other.
"""

import json as _json
from datetime import datetime as _datetime
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from arcgis.geoprocessing import DataFile
from ._util import _id_generator, _feature_input, _set_context, _create_output_service

_log = _logging.getLogger(__name__)

# url = "https://dev003153.esri.com/gax/rest/services/System/GeoAnalyticsTools/GPServer"

_use_async = True

def build_multivariable_grid(input_layers,
                           variable_calculations,
                           bin_size,
                           bin_unit="Meters",
                           bin_type="Square",
                           output_name=None,
                           gis=None):
    """
    Only available at ArcGIS Enterprise 10.6.1 and later.

    The Build Multi-Variable Grid task works with one or more layers of point, line, or polygon
    features. The task generates a grid of square or hexagonal bins and compiles information about
    each input layer into each bin. For each input layer, this information can include the following
    variables:

        + Distance to Nearest - The distance from each bin to the nearest feature.
        + Attribute of Nearest - An attribute value of the feature nearest to each bin.
        + Attribute Summary of Related - A statistical summary of all features within
          search distance of each bin.

    Only variables you specify in variable_calculations will be included in the result layer. These
    variables can help you understand the proximity of your data throughout the extent of your
    analysis. The results can help you answer questions such as the following:

        + Given multiple layers of public transportation infrastructure, where in the city is least
          accessible by public transportation?
        + Given layers of lakes and rivers, what is the name of the water body closest to each
          location in the US?
        + Given a layer of household income, where in the US is the variation of income in the
          surrounding 50 miles the largest?

    The result of Build Multi-Variable Grid can also be used in prediction and classification
    workflows. The task allows you to calculate and compile information from many different data
    sources into a single, spatially continuous layer in one step, reducing the amount of effort
    required to build prediction and classification models.


    =======================   ===================================================================
    **Arguments**             **Description**
    -----------------------   -------------------------------------------------------------------
    input_layers              Required list of FeatureLayers. A list of input layers that will be
                              used in analysis. Each input layer follows the same formatting as
                              described in the Feature Input topic. This can be one of the
                              following:

                                 - A URL to a feature service layer with an optional filter to
                                   select specific features
                                 - A URL to a big data catalog service layer with an optional
                                   filter to select specific features
                                 - A feature collection

    -----------------------   -------------------------------------------------------------------
    variable_calculations     Required list of dictionaries. A JSON array containing objects that
                              describe the input layers and the attributes that will be
                              calculated for each layer.
    -----------------------   -------------------------------------------------------------------
    bin_size                  Required float. The distance for the bins of type binType in the
                              output polygon layer. Enrichment attributes will be calculated at
                              the center of each bin. When generating bins, for Square, the
                              number and units specified determine the height and length of the
                              square. For Hexagon, the number and units specified determine the
                              distance between parallel sides.
    -----------------------   -------------------------------------------------------------------
    bin_unit                  Optional string. The distance unit for the bins that will be used
                              to calculate enrichment attributes.

                              Values: Meters (default), Kilometers, Feet, Miles, NauticalMiles,
                              or Yards
    -----------------------   -------------------------------------------------------------------
    bin_type                  Optional string. The type of bin that will be generated. Bin
                              options are the following:

                                + Hexagon.
                                + Square (default)
    -----------------------   -------------------------------------------------------------------
    output_name               Optional string. output name of the layer
    -----------------------   -------------------------------------------------------------------
    gis                       Optional GIS.  The enterprise site that you want to connect to.
    =======================   ===================================================================

    :returns: Feature Layer

    """
    kwargs=locals()

    gis=_arcgis.env.active_gis if gis is None else gis
    url=gis.properties.helperServices.geoanalytics.url

    params={}
    for key, value in kwargs.items():
        if key == 'variable_calculations':
            import json
            params[key] = json.dumps(value)
        elif value is not None:
            params[key]=value

    if output_name is None:
        output_service_name='Build Multi Variable Grid_' + _id_generator()
        output_name=output_service_name.replace(' ', '_')
    else:
        output_service_name=output_name.replace(' ', '_')

    output_service=_create_output_service(gis, output_name, output_service_name, 'Build Multi Variable Grid ')

    params['output_name']=_json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db={
        "input_layers": (_FeatureSet, "inputLayers"),
        "variable_calculations" : (str, "variableCalculations"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_unit": (str, "binSizeUnit"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values=[
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "BuildMultiVariableGrid", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise


def aggregate_points(point_layer,
                     bin_type = None,
                     bin_size = None,
                     bin_size_unit = None,
                     polygon_layer = None,
                     time_step_interval = None,
                     time_step_interval_unit = None,
                     time_step_repeat_interval = None,
                     time_step_repeat_interval_unit = None,
                     time_step_reference = None,
                     summary_fields = None,
                     output_name = None,
                     gis=None):
    """
    Using a layer of point features and either a layer of area features or bins defined by a specified distance, this tool determines which points fall within each area or bin and calculates statistics about all the points within each area or bin. You may optionally apply time slicing with this tool.

    For example

    * Given point locations of crime incidents, count the number of crimes per county or other administrative district.

    * Find the highest and lowest monthly revenues for franchise locations using 100 km bins.

    This tool works with a layer of point features and a layer of areas features. Input area features can be from a polygon layer or they can be square or hexagonal bins calculated when the tool is run. The tool first determines which points fall within each specified area. After determining this point-in-area spatial relationship, statistics about all points in the area are calculated and assigned to the area. The most basic statistic is the count of the number of points within the area, but you can get other statistics as well.

    For example, suppose you have point features of coffee shop locations and area features of counties, and you want to summarize coffee sales by county. Assuming the coffee shops have a TOTAL_SALES attribute, you can get the sum of all TOTAL_SALES within each county, the minimum or maximum TOTAL_SALES within each county, or other statistics such as the count, range, standard deviation, and variance.

    This tool can also work with data that is time-enabled. If time is enabled on the input points, then the time slicing options are available. Time slicing allows you to calculate the point-in-area relationship while looking at a specific slice in time. For example, you could look at hourly intervals, which would result in outputs for each hour.

    For an example with time, suppose you had point features of every transaction made at various coffee shop locations and no area layer. The data has been recorded over a year and each transaction has a location and a time stamp. Assuming each transaction has a TOTAL_SALES attribute, you can get the sum of all TOTAL_SALES within the space and time of interest. If these transactions are for a single city, we could generate areas that are 1-kilometer grids and look at weekly time slices to summarize the transactions in both time and space.


    ====================================     ====================================================================
    **Argument**                             **Description**
    ------------------------------------     --------------------------------------------------------------------
    point_layer                              Required Input Points layer (features).
    ------------------------------------     --------------------------------------------------------------------
    bin_type                                 Optional string parameter. If polygon_layer is not defined, it is required.
                                             Choice list:['Square', 'Hexagon']
    ------------------------------------     --------------------------------------------------------------------
    bin_size                                 Bin Size (float). Optional parameter.
    ------------------------------------     --------------------------------------------------------------------
    bin_size_unit                            Bin Size Unit (str). Optional parameter.
                                             Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']
    ------------------------------------     --------------------------------------------------------------------
    polygon_layer                            Optional Input Polygons layer (features). If bin_type and bin properties are not defined, it is
                                             required.
    ------------------------------------     --------------------------------------------------------------------
    time_step_interval                       Time Step Interval (int). Optional parameter.
    ------------------------------------     --------------------------------------------------------------------
    time_step_interval_unit                  Time Step Interval Unit (str). Optional parameter.
                                             Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']
    ------------------------------------     --------------------------------------------------------------------
    time_step_repeat_interval                Time Step Repeat Interval (int). Optional parameter.
    ------------------------------------     --------------------------------------------------------------------
    time_step_repeat_interval_unit           Time Step Repeat Interval Unit (str). Optional parameter.
                                             Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']
    ------------------------------------     --------------------------------------------------------------------
    time_step_reference                      Time Step Reference (datetime). Optional parameter.
    ------------------------------------     --------------------------------------------------------------------
    summary_fields                           Summary Statistics (str/list). Optional parameter.

                                             The summary_fields string must enclose a Python list. Each list item must be a Python dictionary
                                             with two keys. See the Key:Value definitions below.

                                             See URL 1 below for full details.
    ------------------------------------     --------------------------------------------------------------------
    output_name                              Output Features Name (str). Optional parameter.
    ------------------------------------     --------------------------------------------------------------------
    gis                                      Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ====================================     ====================================================================


    *Key:Value Dictionary Options for Argument summary_fields*


    =================  =====================================================================
    **Key**            **Value**
    -----------------  ---------------------------------------------------------------------
    statisticType      Required string. Indicates statistic to summarize. See URL 1 below for full explanation.

                       Choice list numeric fields:['Count', 'Sum', 'Mean', 'Min', 'Max', 'Range', 'Stddev', 'Var']

                       Choice list for string fields:['Count', 'Any']
    -----------------  ---------------------------------------------------------------------
    onStatisticField   Required string. Provides the field name to summarize.

                       See https://developers.arcgis.com/python/guide/working-with-feature-layers-and-features/#Querying-feature-layers
                       for instructions to query a feature layer for field names.
    =================  =====================================================================


    For detailed explanation see:

    URL 1: http://resources.arcgis.com/en/help/arcgis-rest-api/index.html#/Aggregate_Points/02r3000002rr000000/

        **Returns:** Output Features as Item

    *Example*

    .. code-block:: python

            # Usage Example: Using summary_fields on a layer.

            agg_pts_item = aggregate_points(input_points_layer,
                              bin_size=0.5,
                              bin_type='Hexagon',
                              bin_size_unit='Miles',
                              summary_fields=[{"statisticType": "Count", "onStatisticField": "fieldName1"}, {"statisticType": "Any", "onStatisticField": "fieldName2"}]
                              )
    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Aggregate Points Analysis_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Aggregate Points')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})
    if isinstance(summary_fields, list):
        import json
        summary_fields = json.dumps(summary_fields)
    _set_context(params)

    param_db = {
        "point_layer": (_FeatureSet, "pointLayer"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "polygon_layer": (_FeatureSet, "polygonLayer"),
        "time_step_interval": (int, "timeStepInterval"),
        "time_step_interval_unit": (str, "timeStepIntervalUnit"),
        "time_step_repeat_interval": (int, "timeStepRepeatInterval"),
        "time_step_repeat_interval_unit": (str, "timeStepRepeatIntervalUnit"),
        "time_step_reference": (_datetime, "timeStepReference"),
        "summary_fields": (str, "summaryFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "AggregatePoints", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

aggregate_points.__annotations__ = {
                     'bin_type': str,
                     'bin_size': float,
                     'bin_size_unit': str,
                     'time_step_interval': int,
                     'time_step_interval_unit': str,
                     'time_step_repeat_interval': int,
                     'time_step_repeat_interval_unit': str,
                     'time_step_reference': _datetime,
                     'summary_fields': str,
                     'output_name': str
                }

def _describe_dataset(input_layer,
                     gis=None):
    """


Parameters:

   input_layer: Input Dataset (feature layer). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output JSON as a str


    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "context": (str, "context"),
        "output": (str, "Output JSON"),
    }
    return_values = [
        {"name": "output", "display_name": "Output JSON", "type": str},
    ]

    return _execute_gp_tool(gis, "DescribeDataset", params, param_db, return_values, _use_async, url, True)

_describe_dataset.__annotations__ = {'return': str}

def join_features(target_layer,
                  join_layer,
                  join_operation = """JoinOneToOne""",
                  join_fields = None,
                  summary_fields = None,
                  spatial_relationship = None,
                  spatial_near_distance = None,
                  spatial_near_distance_unit = None,
                  temporal_relationship = None,
                  temporal_near_distance = None,
                  temporal_near_distance_unit = None,
                  attribute_relationship = None,
                  join_condition = None,
                  output_name = None,
                  gis=None):
    """
    Using either feature layers or tabular data, you can join features and records based on specific relationships between the input layers or tables. Joins will be determined by spatial, temporal, and attribute relationships, and summary statistics can be optionally calculated.

    For example

    * Given point locations of crime incidents with a time, join the crime data to itself specifying a spatial relationship of crimes within 1 kilometer of each other and that occurred within 1 hour of each other to determine if there are a sequence of crimes close to each other in space and time.

    * Given a table of ZIP Codes with demographic information and area features representing residential buildings, join the demographic information to the residences so each residence now has the information.

    The Join Features task works with two layers. Join Features joins attributes from one feature to another based on spatial, temporal, and attribute relationships or some combination of the three. The tool determines all input features that meet the specified join conditions and joins the second input layer to the first. You can optionally join all features to the matching features or summarize the matching features.

    Join Features can be applied to points, lines, areas, and tables. A temporal join requires that your input data is time-enabled, and a spatial join requires that your data has a geometry.



Parameters:

   target_layer: Target Features (feature input). Required parameter.

   join_layer: Join Features (feature input). Required parameter.

   join_operation: Join Operation (str). Required parameter.
      Choice list:['JoinOneToOne', 'JoinOneToMany']

   join_fields: Join Fields (str). Optional parameter.

   summary_fields: Summary Statistics (str). Optional parameter.

   spatial_relationship: Spatial Relationship (str). Optional parameter.
      Choice list:['Equals', 'Intersects', 'Contains', 'Within', 'Crosses', 'Touches', 'Overlaps', 'Near']

   spatial_near_distance: Near Spatial Distance (float). Optional parameter.

   spatial_near_distance_unit: Near Spatial Distance Unit (str). Optional parameter.
      Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

   temporal_relationship: Temporal Relationship (str). Optional parameter.
      Choice list:['Equals', 'Intersects', 'During', 'Contains', 'Finishes', 'FinishedBy', 'Meets', 'MetBy', 'Overlaps', 'OverlappedBy', 'Starts', 'StartedBy', 'Near']

   temporal_near_distance: Near Temporal Distance (int). Optional parameter.

   temporal_near_distance_unit: Near Temporal Distance Unit (str). Optional parameter.
      Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

   attribute_relationship: Attribute Relationships (str). Optional parameter.

   join_condition: Join Condition (str). Optional parameter.

   output_name: Output Features Name (str). Optional parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Features as Feature Layer Collection Item


    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Join Features Analysis_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Join Features')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "target_layer": (_FeatureSet, "targetLayer"),
        "join_layer": (_FeatureSet, "joinLayer"),
        "join_operation": (str, "joinOperation"),
        "join_fields": (str, "joinFields"),
        "summary_fields": (str, "summaryFields"),
        "spatial_relationship": (str, "spatialRelationship"),
        "spatial_near_distance": (float, "spatialNearDistance"),
        "spatial_near_distance_unit": (str, "spatialNearDistanceUnit"),
        "temporal_relationship": (str, "temporalRelationship"),
        "temporal_near_distance": (int, "temporalNearDistance"),
        "temporal_near_distance_unit": (str, "temporalNearDistanceUnit"),
        "attribute_relationship": (str, "attributeRelationship"),
        "join_condition": (str, "joinCondition"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "JoinFeatures", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

join_features.__annotations__ = {
                  'join_operation': str,
                  'join_fields': str,
                  'summary_fields': str,
                  'spatial_relationship': str,
                  'spatial_near_distance': float,
                  'spatial_near_distance_unit': str,
                  'temporal_relationship': str,
                  'temporal_near_distance': int,
                  'temporal_near_distance_unit': str,
                  'attribute_relationship': str,
                  'join_condition': str,
                  'output_name': str}

def reconstruct_tracks(input_layer,
                       track_fields,
                       method = """Planar""",
                       buffer_field = None,
                       summary_fields = None,
                       time_split = None,
                       time_split_unit = None,
                       distance_split=None,
                       distance_split_unit=None,
                       output_name = None,
                       gis=None):
    """

    Using a time-enabled layer of point or polygon features that represent an instant in time, this tool determines which input features belong in a track and will order the inputs sequentially in time. Statistics are optionally calculated for the input features within each track.

    For example

    * Given point locations and time of hurricane measurements, calculate the mean wind speed and max wind pressure of the hurricane.

    * Find the highest and lowest monthly revenues for franchise locations using 100 km bins.

        This tool works with a time-enabled layer of either point or polygon features that represent an instant in time. It first determines which features belong to a track using an identifier. Using the time at each location, the tracks are ordered sequentially and transformed into a line or polygon representing the path of movement over time. Optionally, the input may be buffered by a field, which will create a polygon at each location. These buffered points, or if the inputs are polygons, are then joined sequentially to create a track as a polygon where the width is representative of the attribute of interest. Resulting tracks have a start and end time, which represent temporally the first and last feature in a given track. When the tracks are created, statistics about the input features are calculated and assigned to the output track. The most basic statistic is the count of points within the area, but other statistics can be calculated as well.

    Features in time-enabled layers can be represented in one of two ways:

    Instant-A single moment in time
    Interval-A start and end time
    For example, suppose you have GPS measurements of hurricanes every 10 minutes. Each GPS measurement records the hurricane's name, location, time of recording, and wind speed. With this information, you could create tracks for each hurricane using the name for track identification, and tracks for each hurricane would be generated. Additionally, you could calculate statistics such as the mean, max, and minimum wind speed of each hurricane, as well as the count of measurements within each track.

    Using the same example, you could buffer your tracks by the wind speed. This would buffer each measurement by the wind speed field at that location, and join the buffered areas together, creating a polygon representative of the track path, as well as the changes in wind speed as the hurricanes progressed.

    Parameters:

   input_layer: Input Features (feature input). Required parameter.

   track_fields: Track Fields (str). Required parameter.

   method: Method (str). Required parameter.
      Choice list:['Geodesic', 'Planar']

   buffer_field: Buffer Distance Field (str). Optional parameter.

   summary_fields: Summary Statistics (str/list). Optional parameter.

   time_split: Duration Split Threshold (int). Optional parameter.

   time_split_unit: Duration Split Threshold Unit (str). Optional parameter.
      Choice list:['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds']

   distance_split: A distance used to split tracks. Any features in the inputLayer that are in the same track and are greater than this distance apart will be split into a new track. The units of the distance values are supplied by the distance_unit parameter.

   distance_split_unit: The distance unit to be used with the distance value specified in distanceSplit.
       Values: Meters,Kilometers,Feet,Miles,NauticalMiles, or Yards

   output_name: Output Features Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Features as a Feature Layer Collection Item


    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Reconstructed Tracks_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Reconstruct Tracks')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    if isinstance(summary_fields, list):
        import json
        summary_fields = json.dumps(summary_fields)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "track_fields": (str, "trackFields"),
        "method": (str, "method"),
        "buffer_field": (str, "bufferField"),
        "summary_fields": (str, "summaryFields"),
        "time_split": (int, "timeSplit"),
        "time_split_unit": (str, "timeSplitUnit"),
        "distance_split": (int, "distanceSplit"),
        "distance_split_unit": (str, "distanceSplitUnit"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, "ReconstructTracks", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise


reconstruct_tracks.__annotations__ = {
                       'track_fields':str,
                       'method': str,
                       'buffer_field': str,
                       'summary_fields': str,
                       'time_split': int,
                       'time_split_unit': str,
                       'output_name': str}

def summarize_attributes(input_layer,
                         fields = None,
                         summary_fields = None,
                         output_name = None,
                         gis=None):
    """
    Using either feature or tabular data, this tool summarizes statistics for specified fields.

    For example

    * Given locations of grocery stores with a field COMPANY_NAME, summarize the stores by the company name to determine statistics for each company.

    * Given a table of grocery stores with fields COMPANY_NAME and COUNTY, summarize the stores by the company name and county to determine statistics for each company within each county.

    This tool summarizes all the matching values in one or more fields and calculates statistics on them. The most basic statistic is the count of features that have been summarized together, but you can calculate more advanced statistics as well.

    For example, suppose you have point features of store locations with a field representing the DISTRICT_MANAGER_NAME and you want to summarize coffee sales by manager. You can specify the field DISTRICT_MANAGER_NAME as the field to dissolve on, and all rows of data representing individual managers will be summarized. This means all store locations that are managed by Manager1 will be summarized into one row with summary statistics calculated. In this instance, statistics like the count of the number of stores and the sum of TOTAL_SALES for all stores that Manager1 manages would be calculated as well as for any other manager listed in the DISTRICT_MANAGER_NAME field.

    Parameters:

   input_layer: Input Features (feature input). Required parameter.

   fields: Summary Fields (str). Required parameter.

   summary_fields: Summary Statistics (str/list). Optional parameter.

   output_name: Output Features Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Features as a _FeatureSet


    """
    kwargs = locals()


    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Summarize Attributes_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Summarize Attributes')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    if isinstance(summary_fields, list):
        import json
        summary_fields = json.dumps(summary_fields)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "fields": (str, "fields"),
        "summary_fields": (str, "summaryFields"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, "SummarizeAttributes", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

summarize_attributes.__annotations__ = {
                         'fields': str,
                         'summary_fields': str,
                         'output_name': str}

def summarize_within(summary_polygons,
                     summarized_layer,
                     bin_type = None,
                     bin_size = None,
                     bin_size_unit = None,
                     standard_summary_fields = None,
                     weighted_summary_fields = None,
                     sum_shape = True,
                     shape_units = None,
                     output_name = None,
                     gis=None):
    """
    Finds areas (and portions of areas) that overlap between two layers and calculates statistics about the overlap.

    For example

    * Given a layer of watershed areas and a layer of land-use areas by land-use type, calculate total acreage of land-use type for each watershed.

    * Given a layer of parcels in a county and a layer of city boundaries, summarize the average value of vacant parcels within each city.




   Parameters:

   summary_polygons: Summary Polygons Layer (feature input). Required parameter.

   summarized_layer: Layer To Summarize (feature input). Required parameter.

   bin_type: Output Bin Type (str). Optional parameter.
      Choice list:['Square', 'Hexagon']

   bin_size: Output Bin Size (float). Optional parameter.

   bin_size_unit: Output Bin Size Unit (str). Optional parameter.
      Choice list:['Feet', 'Yards', 'Miles', 'Meters', 'Kilometers', 'NauticalMiles']

   standard_summary_fields: Unweighted Summary Statistics (str). Optional parameter.

   weighted_summary_fields: Proportional Summary Statistics (str). Optional parameter.

   sum_shape: Summarize Shape (bool). Optional parameter.

   shape_units: Shape Measure Output Unit (str). Optional parameter.
      Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'SquareMeters', 'SquareKilometers', 'Hectares', 'SquareFeet', 'SquareYards', 'SquareMiles', 'Acres']

   output_name: Output Features Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Features as a layer Item


    """
    kwargs = locals()


    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Summarize Within_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Summarize Within')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "summary_polygons": (_FeatureSet, "summaryPolygons"),
        "bin_type": (str, "binType"),
        "bin_size": (float, "binSize"),
        "bin_size_unit": (str, "binSizeUnit"),
        "summarized_layer": (_FeatureSet, "summarizedLayer"),
        "standard_summary_fields": (str, "standardSummaryFields"),
        "weighted_summary_fields": (str, "weightedSummaryFields"),
        "sum_shape": (bool, "sumShape"),
        "shape_units": (str, "shapeUnits"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Features"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        _execute_gp_tool(gis, "SummarizeWithin", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

summarize_within.__annotations__ = {
                     'bin_type': str,
                     'bin_size': float,
                     'bin_size_unit': str,
                     'standard_summary_fields': str,
                     'weighted_summary_fields': str,
                     'sum_shape': bool,
                     'shape_units': str,
                     'output_name': str
                }




""" Defines store functions for working with Workers.
"""

from ... import workforce
from .utils import add_features, update_features, remove_features, validate


def get_dispatcher(project, object_id=None, global_id=None, user_id=None):
    """ Gets the identified Dispatcher.  Exactly one form of identification should be provided.
        :param project:
        :param object_id: The dispatcher's OBJECTID.
        :param global_id: The dispatcher's GlobalID.
        :param user_id: The dispatcher's named user user_id.
        :returns: Dispatcher
    """
    if object_id:
        where = "{} = {}".format(project._dispatcher_schema.object_id, object_id)
    elif global_id:
        where = "{} = '{}'".format(project._dispatcher_schema.global_id, global_id)
    elif user_id:
        where = "{} = '{}'".format(project._dispatcher_schema.user_id, user_id)
    else:
        where = "1=0"
    dispatchers = query_dispatchers(project, where)
    return dispatchers[0] if dispatchers else None


def get_dispatchers(project):
    """ Gets all Dispatchers in the project.
        :param project:
        :returns: list of Dispatchers
    """
    return query_dispatchers(project, '1=1')


def query_dispatchers(project, where):
    """ Executes a query against the dispatchers feature layer.
        :param project: The project in which to query dispatchers.
        :param where: An ArcGIS where clause.
        :returns: list of Dispatchers
    """
    features = project.dispatchers_layer.query(where, return_all_freatures=True).features
    return [workforce.Dispatcher(project, feature) for feature in features]


def add_dispatcher(project, feature=None, contact_number=None, name=None, user_id=None):
    """
    Adds a new dispatcher to the project
    """
    project._update_cached_objects()
    dispatcher = workforce.Dispatcher(project,
                                      feature,
                                      contact_number,
                                      name,
                                      user_id
                                      )
    return add_dispatchers(project, [dispatcher])[0]


def update_dispatcher(project, dispatcher, contact_number=None, name=None, user_id=None):
    """
        Updates a dispatcher and submits changes to the server
    """
    project._update_cached_objects()
    if contact_number:
        dispatcher.contact_number = contact_number
    if name:
        dispatcher.name = name
    if user_id:
        dispatcher.user_id = user_id
    return update_dispatchers(project, [dispatcher])[0]


def add_dispatchers(project, dispatchers):
    """ Adds Dispatchers to a project.

        Side effect: Upon successful addition on the server, the object_id and global_id fields of
        each Dispatcher in dispatchers will be updated to the values assigned by the server.

        :param project:
        :param dispatchers: list of Dispatchers
        :raises ValidationError: Indicates that one or more dispatchers failed validation.
        :raises ServerError: Indicates that the server rejected the dispatchers.
    """
    project._update_cached_objects()
    if dispatchers:
        use_global_ids = True
        for dispatcher in dispatchers:
            validate(dispatcher._validate_for_add)

            if dispatcher.global_id is None:
                use_global_ids = False

        features = [dispatcher.feature for dispatcher in dispatchers]
        add_features(project.dispatchers_layer, features, use_global_ids)

    # add dispatcher named users to the project's group.
    project.group.add_users([dispatcher.user_id for dispatcher in dispatchers])
    return dispatchers


def update_dispatchers(project, dispatchers):
    """ Updates Dispatchers.
        :param project:
        :param dispatchers: list of Dispatchers to update
        :raises ValidationError: Indicates that one or more dispatchers failed validation.
        :raises ServerError: Indicates that the server rejected the dispatchers.
    """
    project._update_cached_objects()
    if dispatchers:
        for dispatcher in dispatchers:
            validate(dispatcher._validate_for_update)
        features = [dispatcher.feature for dispatcher in dispatchers]
        update_features(project.dispatchers_layer, features)
    return dispatchers


def delete_dispatchers(project, dispatchers):
    """ Removes Dispatchers from the project.
        :param project:
        :param dispatchers: list of Dispatchers
        :raises ValidationError: Indicates that one or more dispatchers failed validation.
        :raises ServerError: Indicates that the server rejected the removal.
    """
    project._update_cached_objects()
    if dispatchers:
        for dispatcher in dispatchers:
            validate(dispatcher._validate_for_remove)
        features = [dispatcher.feature for dispatcher in dispatchers]
        remove_features(project.dispatchers_layer, features)

        # Remove dispatcher named users from the project's group, unless they are also workers.
        user_ids = [dispatcher.user_id for dispatcher in dispatchers]
        where = "{} in ({})".format(project._worker_schema.user_id,
                                    ','.join(["'{}'".format(user_id) for user_id in user_ids]))
        workers = workforce._store.query_workers(project, where)
        for worker in workers:
            user_ids.remove(worker.user_id)
        if user_ids:
            project.group.remove_users(user_ids)
